'''
.. autoclass:: Composition
   :members:
   :special-members:
'''
import collections

class Composition(object):
    'a single composition, possibly with other compositions inside it'
    def __init__(self, *items):
        'initialize this Composition by _wrapping each item in a list'
        self.funcs = [
            self._wrap(item, splat=False) for item
            in list(items)
        ]

    def _flattened(self):
        'a _flattened version of this composition'
        flattened = []
        for item, splat in self.funcs:
            if isinstance(item, Composition):
                flattened += item._flattened()

            else:
                flattened.append((item, splat))

        return flattened

    def _wrap(self, func, splat=False):
        '''
        Functions are kept track of in a tuple of (function, splat). This
        function knows about this format, and makes items adhere to it. If an
        item is already a tuple, it will be ignored.
        '''
        if isinstance(func, tuple):
            return func

        return (func, splat)

    def __gt__(self, other):
        '''
        Use ``>`` to append to the composition, without splatting.
        '''
        return Composition(self, self._wrap(other, splat=False))

    def __rshift__(self, other):
        '''
        Use ``>>`` to append to the composition, with splatting.
        '''
        return Composition(self, self._wrap(other, splat=True))

    def __call__(self, *args, **kwargs):
        '''
        Call has two basic steps: getting an initial value, then running that
        initial value through each step of the composition.

        Note that the initial function will *always* receive ``*args`` and
        ``**kwargs``. All subsequent calls respect the second pair of the
        tuple, which will splat the output of the previous function (using the
        form ``func(*result)``.)
        '''
        result = self.funcs[0][0](*args, **kwargs)

        for func, splat in self.funcs[1:]:
            if splat:
                result = func(*result)
            else:
                result = func(result)

        return result



'test composition'
from unittest import TestCase

from aaron.composition import Composition

# some example functions
def add_one(n):
    return n + 1

def add_two(n):
    return n + 2

def ints_less_than(n):
    return range(1, n)

def product(*ns):
    result = 1
    for n in ns:
        result *= n

    return result

# test
class WrapTests(TestCase):
    def test__wraps(self):
        self.assertEqual(
            Composition()._wrap('x'),
            ('x', False)
        )

    def test__wraps_splat(self):
        self.assertEqual(
            Composition()._wrap('x', splat=True),
            ('x', True)
        )

    def test_ignores_tuples(self):
        self.assertEqual(
            Composition()._wrap(('1', '2')),
            ('1', '2')
        )

class GtTests(TestCase):
    def test_gt(self):
        comp = Composition(add_one) > add_two

        self.assertEqual(
            [
                (add_one, False),
                (add_two, False)
            ],
            comp._flattened()
        )

class RshiftTest(TestCase):
    def test_rshift(self):
        comp = Composition(ints_less_than) >> product

        self.assertEqual(
            [
                (ints_less_than, False),
                (product, True)
            ],
            comp._flattened()
        )

class CallTest(TestCase):
    def test_nosplat(self):
        comp = Composition(add_one) > add_two

        self.assertEqual(comp(0), 3)

    def test_splat(self):
        comp = Composition(ints_less_than) >> product

        self.assertEqual(comp(4), 6)

    def test_short(self):
        comp = Composition(add_one)

        self.assertEqual(comp(0), 1)

    def test_long(self):
        comp = Composition(add_one)
        for _ in range(99):
            comp = comp > add_one

        self.assertEqual(comp(0), 100)



'tests for decorators'
from unittest import TestCase

from aaron.composition import Composition
from aaron.decorators import composable

def add_one(n):
    return n + 1

class ComposableTests(TestCase):
    'tests for composable decorator'

    def test_isinstance(self):
        self.assertFalse(isinstance(add_one, Composition))
        self.assertTrue(isinstance(composable(add_one), Composition))
